using Tensorial: fromvoigt, ⊡, ⋅

import Tensorial: contraction, double_contraction

export isequivalent, ⩵

TensorStress(σ::EngineeringStress) = TensorStress(σ[1], σ[6], σ[5], σ[2], σ[4], σ[3])

EngineeringStress(σ::TensorStress) =
    EngineeringStress(σ[1, 1], σ[2, 2], σ[3, 3], σ[2, 3], σ[1, 3], σ[1, 2])

TensorStrain(ϵ::EngineeringStrain) =
    TensorStrain(ϵ[1], ϵ[6] / 2, ϵ[5] / 2, ϵ[2], ϵ[4] / 2, ϵ[3])

EngineeringStrain(ε::TensorStrain) =
    EngineeringStrain(ε[1, 1], ε[2, 2], ε[3, 3], 2ε[2, 3], 2ε[1, 3], 2ε[1, 2])

StiffnessMatrix(c::StiffnessTensor) = StiffnessMatrix(
    c[1, 1, 1, 1],
    c[1, 1, 2, 2],
    c[1, 1, 3, 3],
    c[1, 1, 2, 3],
    c[1, 1, 1, 3],
    c[1, 1, 1, 2],
    c[2, 2, 2, 2],
    c[2, 2, 3, 3],
    c[2, 2, 2, 3],
    c[2, 2, 1, 3],
    c[2, 2, 1, 2],
    c[3, 3, 3, 3],
    c[3, 3, 2, 3],
    c[3, 3, 1, 3],
    c[3, 3, 1, 2],
    c[2, 3, 2, 3],
    c[2, 3, 1, 3],
    c[2, 3, 1, 2],
    c[1, 3, 1, 3],
    c[1, 3, 1, 2],
    c[1, 2, 1, 2],
)

StiffnessTensor(c::StiffnessMatrix) =
    StiffnessTensor(fromvoigt(SymmetricFourthOrderTensor{3}, c.data))

ComplianceMatrix(s::ComplianceTensor) = ComplianceMatrix(
    s[1, 1, 1, 1],
    s[1, 1, 2, 2],
    s[1, 1, 3, 3],
    2s[1, 1, 2, 3],
    2s[1, 1, 1, 3],
    2s[1, 1, 1, 2],
    s[2, 2, 2, 2],
    s[2, 2, 3, 3],
    2s[2, 2, 2, 3],
    2s[2, 2, 1, 3],
    2s[2, 2, 1, 2],
    s[3, 3, 3, 3],
    2s[3, 3, 2, 3],
    2s[3, 3, 1, 3],
    2s[3, 3, 1, 2],
    4s[2, 3, 2, 3],
    4s[2, 3, 1, 3],
    4s[2, 3, 1, 2],
    4s[1, 3, 1, 3],
    4s[1, 3, 1, 2],
    4s[1, 2, 1, 2],
)

ComplianceTensor(s::ComplianceMatrix) = ComplianceTensor(
    SymmetricFourthOrderTensor{3}(function (i, j, k, l)
        if i == j && k == l
            return s[i, k]
        elseif i != j && k != l  # 4 = 9 - (2+3), 5 = 9 - (1+3), 6 = 9 - (1+2)
            return s[9 - (i + j), 9 - (k + l)] / 4
        elseif i == j && k != l
            return s[i, 9 - (k + l)] / 2
        else  # i != j && k == l
            return s[9 - (i + j), k] / 2
        end
    end)
)

Base.:*(c::StiffnessMatrix, ϵ::EngineeringStrain) = EngineeringStress(c.data ⋅ ϵ.data)
Base.:*(s::ComplianceMatrix, σ::EngineeringStress) = EngineeringStrain(s.data ⋅ σ.data)

Base.inv(c::StiffnessTensor) = ComplianceTensor(inv(c.data))
Base.inv(s::ComplianceTensor) = StiffnessTensor(inv(s.data))
Base.inv(c::StiffnessMatrix) = ComplianceMatrix(inv(c.data))
Base.inv(s::ComplianceMatrix) = StiffnessMatrix(inv(s.data))

contraction(c::StiffnessTensor, ε::TensorStrain, ::Val{2}) = TensorStress(c.data ⊡ ε.data)
contraction(s::ComplianceTensor, σ::TensorStress, ::Val{2}) = TensorStrain(s.data ⊡ σ.data)

@inline double_contraction(c::StiffnessTensor, ε::TensorStrain) = contraction(c, ε, Val(2))
@inline double_contraction(s::ComplianceTensor, σ::TensorStress) = contraction(s, σ, Val(2))

# See https://discourse.julialang.org/t/how-to-compare-two-vectors-whose-elements-are-equal-but-their-types-are-not-the-same/94309/6
for (S, T) in (
    (:EngineeringStress, :EngineeringStrain),
    (:TensorStress, :TensorStrain),
    (:StiffnessTensor, :ComplianceTensor),
    (:StiffnessMatrix, :ComplianceMatrix),
)
    @eval begin
        isequivalent(s::$S, t::$T) = false
        isequivalent(t::$T, s::$S) = isequivalent(s, t)
    end
end
isequivalent(x, y) = x == y  # Fallback
const ⩵ = isequivalent
